@@ -32,7 +32,7 @@ class DocumentOtherLoadedImpl : public DocumentImpl {
 public:
 	DocumentOtherLoadedImpl(Document* document)
 	: DocumentImpl(document) {
-		setImage(QImage(), false);
+		setImage(QImage());
 		setImageFormat(0);
 	}
 
@@ -42,6 +42,8 @@ public:
 
 	virtual MimeTypeUtils::Kind urlKind() const { return MimeTypeUtils::KIND_FILE; }
 
+	virtual bool canBeSaved() const { return false; }
+
 	virtual int duration() const;
 };
 
