@@ -45,6 +45,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "fileviewconfig.h"
 #include "filethumbnailviewitem.h"
 #include "archive.h"
+#include "dragpixmapgenerator.h"
 #include "thumbnailloadjob.h"
 #include "busylevelmanager.h"
 #include "imageloader.h"
@@ -69,7 +70,6 @@ static const int THUMBNAIL_UPDATE_DELAY=500;
 static const int RIGHT_TEXT_WIDTH=128;
 static const int BOTTOM_MIN_TEXT_WIDTH=96;
 
-
 class ProgressWidget : public QFrame {
 	KProgress* mProgressBar;
 	QPushButton* mStop;
@@ -484,7 +484,7 @@ void FileThumbnailView::setSelected(const KFileItem* fileItem,bool enable) {
 	if (!fileItem) return;
 
 	FileThumbnailViewItem* iconItem=viewItem(this, fileItem);
-	if (iconItem) QIconView::setSelected(iconItem,enable);
+	if (iconItem) QIconView::setSelected(iconItem, enable, true /* do not unselect others */);
 }
 
 
@@ -773,11 +773,71 @@ void FileThumbnailView::prefetchDone() {
 //
 //--------------------------------------------------------------------------
 void FileThumbnailView::startDrag() {
+	/**
+	 * The item drawer for DragPixmapGenerator
+	 */
+	struct ItemDrawer : public DragPixmapItemDrawer<KFileItem*> {
+		ItemDrawer(FileThumbnailView* view)
+		: mView(view) {}
+
+		QSize itemSize(KFileItem* fileItem) {
+			QPixmap* pix = pixmapFromFileItem(fileItem);
+			if (!pix) return QSize();
+					
+			QSize size = pix->size();
+			int maxWidth = mGenerator->maxWidth();
+			if (size.width() > maxWidth) {
+				size.rheight() = size.height() * maxWidth / size.width();
+				size.rwidth() = maxWidth;
+			}
+			return size;
+		}
+
+		int spacing() const {
+			return 2;
+		}
+			
+		void drawItem(QPainter* painter, int left, int top, KFileItem* fileItem) {
+			QPixmap* pix = pixmapFromFileItem(fileItem);
+			if (!pix) return;
+
+			QSize size = itemSize(fileItem);
+			left += (mGenerator->pixmapWidth() - size.width()) / 2;
+			if (size == pix->size()) {
+				painter->drawPixmap(left, top, *pix);
+				return;
+			}
+			
+			QImage img = pix->convertToImage();
+			img = img.smoothScale(size, QImage::ScaleMin);
+			painter->drawImage(left, top, img);
+		}
+
+		QPixmap* pixmapFromFileItem(KFileItem* fileItem) {
+			FileThumbnailViewItem* iconItem = viewItem(mView, fileItem);
+			Q_ASSERT(iconItem);
+			if (!iconItem) return 0;
+			
+			QPixmap* pix = iconItem->pixmap();
+			Q_ASSERT(pix);
+			if (!pix) return 0;
+			return pix;
+		}
+		
+		FileThumbnailView* mView;
+	};
+	ItemDrawer drawer(this);
+
+
 	KURL::List urls;
 	KFileItemListIterator it(*KFileView::selectedItems());
 
+	DragPixmapGenerator<KFileItem*> generator;
+	generator.setItemDrawer(&drawer);
+	
 	for ( ; it.current(); ++it ) {
 		urls.append(it.current()->url());
+		generator.addItem(it.current());
 	}
 
 	if (urls.isEmpty()) {
@@ -786,14 +846,9 @@ void FileThumbnailView::startDrag() {
 	}
 
 	QDragObject* drag=new KURLDrag(urls, this, 0);
-	QPixmap dragPixmap;
-	if (urls.count()>1) {
-		dragPixmap=SmallIcon("kmultiple");
-	} else {
-		dragPixmap=KFileView::selectedItems()->getFirst()->pixmap(16);
-	}
-	drag->setPixmap( dragPixmap, QPoint(dragPixmap.width()/2, dragPixmap.height()/2) );
+	QPixmap dragPixmap = generator.generate();
 
+	drag->setPixmap( dragPixmap, QPoint(generator.DRAG_OFFSET, -generator.DRAG_OFFSET));
 	drag->dragCopy();
 }
 
