@@ -25,28 +25,44 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 // Qt
 #include <qtimer.h>
+#include <qwmatrix.h>
 
 // KDE
 #include <kapplication.h>
+#include <kmimetype.h>
 
 // Local
 #include "cache.h"
+#include "imageutils/imageutils.h"
+#include "imageutils/jpegcontent.h"
 
 #include "imageloader.moc"
 namespace Gwenview {
 
 const unsigned int DECODE_CHUNK_SIZE=4096;
 
+/** Interval between image updates, in milli seconds */
+const int IMAGE_UPDATE_INTERVAL=100;
+
 #undef ENABLE_LOG
 #undef LOG
-//#define ENABLE_LOG
-#ifdef ENABLE_LOG
+#undef LOG2
+
+#define ENABLE_LOG 0
+
+#if ENABLE_LOG >= 1
 #define LOG(x) kdDebug() << k_funcinfo << x << endl
 #else
 #define LOG(x) ;
 #endif
 
-static QMap< KURL, ImageLoader* > loaders;
+#if ENABLE_LOG >= 2
+#define LOG2(x) kdDebug() << k_funcinfo << x << endl
+#else
+#define LOG2(x) ;
+#endif
+
+static QMap< KURL, ImageLoader* > sLoaders;
 
 //---------------------------------------------------------------------
 //
@@ -169,24 +185,45 @@ struct OwnerData {
 	BusyLevel priority;
 };
 
+enum GetState { 
+	GET_PENDING_STAT, // Stat has not been started
+	GET_STATING,      // Stat has been started
+	GET_PENDING_GET,  // Stat is done, get has not been started
+	GET_GETTING,      // Get has been started
+	GET_DONE,         // All data has been received
+};
+
+
+enum DecodeState {
+	DECODE_WAITING,                   // No data to decode yet
+	DECODE_PENDING_THREADED_DECODING, // Waiting for all data to start threaded decoding
+	DECODE_THREADED_DECODING,         // Threaded decoder is running
+	DECODE_INCREMENTAL_DECODING,      // Incremental decoder is running
+	DECODE_INCREMENTAL_DECODING_DONE, // Incremental decoder is done
+	DECODE_CACHED,                    // Image has been obtained from cache, but raw data was missing. Wait for get to finish.
+	DECODE_DONE,                      // All done
+};
+
 class ImageLoaderPrivate {
 public:
 	ImageLoaderPrivate(ImageLoader* impl)
 	: mDecodedSize(0)
-	, mUseThread(false)
+	, mGetState(GET_PENDING_STAT)
+	, mDecodeState(DECODE_WAITING)
 	, mDecoder(impl)
-	, mUpdatedDuringLoad(false)
 	, mSuspended(false)
-	, mGetComplete(false)
-	, mAsyncImageComplete(false)
 	, mNextFrameDelay(0)
 	, mWasFrameData(false)
+	, mOrientation(ImageUtils::NOT_AVAILABLE)
 	, mURLKind(MimeTypeUtils::KIND_UNKNOWN)
-	, mDecodeComplete( false )
-	, mStatPending( false )
-	, mGetPending( false )
 	{}
 
+	// How many of the raw data we have already decoded
+	unsigned int mDecodedSize;
+
+	GetState mGetState;
+	DecodeState mDecodeState;
+
 	KURL mURL;
 
 	// The file timestamp
@@ -194,25 +231,13 @@ public:
 
 	// The raw data we get
 	QByteArray mRawData;
-
-	// How many of the raw data we have already decoded
-	unsigned int mDecodedSize;
 	
-	// Whether we are using a thread to decode the image (if the async decoder
-	// failed)
-	bool mUseThread;
-
 	// The async decoder and it's waking timer	
 	QImageDecoder mDecoder;
 	QTimer mDecoderTimer;
 
 	// The decoder thread
 	DecoderThread mDecoderThread;
-	
-	// Set to true if at least one changed() signals have been emitted
-	bool mUpdatedDuringLoad;
-	// Set to image size if sizeLoaded() signal has been emitted
-	QSize mKnownSize;
 
 	// A rect of recently loaded pixels that the rest of the application has
 	// not been notified about with the imageChanged() signal
@@ -224,12 +249,6 @@ public:
 	// Whether the loading should be suspended
 	bool mSuspended;
 	
-	// Set to true when all the raw data has been received
-	bool mGetComplete;
-	
-	// Set to true when all the image has been decoded
-	bool mAsyncImageComplete;
-
 	// Delay used for next frame after it's finished decoding.
 	int mNextFrameDelay;
 
@@ -242,17 +261,21 @@ public:
 	ImageFrames mFrames;
 
 	QCString mImageFormat;
+
+	ImageUtils::Orientation mOrientation;
 	
+	QString mMimeType;
 	MimeTypeUtils::Kind mURLKind;
 
 	QValueVector< OwnerData > mOwners; // loaders may be shared
-
-	bool mDecodeComplete; // is decoding fully completed?
-
-	// KIO::stat() hasn't been done yet (because of being suspended since the beginning)
-	bool mStatPending;
-	// the same for KIO::get()
-	bool mGetPending;
+	
+	
+	void determineImageFormat() {
+		Q_ASSERT(mRawData.size()>0);
+		QBuffer buffer(mRawData);
+		buffer.open(IO_ReadOnly);
+		mImageFormat = QImageIO::imageFormat(&buffer);
+	}
 };
 
 
@@ -291,22 +314,21 @@ void ImageLoader::startLoading() {
 	connect(&d->mDecoderTimer, SIGNAL(timeout()), this, SLOT(decodeChunk()) );
 
 	connect(&d->mDecoderThread, SIGNAL(succeeded()),
-		this, SLOT(slotImageDecoded()) );
+		this, SLOT(slotDecoderThreadSucceeded()) );
 	connect(&d->mDecoderThread, SIGNAL(failed()),
 		this, SLOT(slotDecoderThreadFailed()) );
 
-	d->mStatPending = true;
 	checkPendingStat();
 }
 
 void ImageLoader::checkPendingStat() {
-	if( d->mSuspended || !d->mStatPending ) return;
+	if( d->mSuspended || d->mGetState != GET_PENDING_STAT ) return;
 
 	KIO::Job* job=KIO::stat( d->mURL, false );
 	job->setWindow(KApplication::kApplication()->mainWidget());
 	connect(job, SIGNAL(result(KIO::Job*)),
 		this, SLOT(slotStatResult(KIO::Job*)) );
-	d->mStatPending = false;
+	d->mGetState = GET_STATING;
 }
 
 void ImageLoader::slotStatResult(KIO::Job* job) {
@@ -325,22 +347,32 @@ void ImageLoader::slotStatResult(KIO::Job* job) {
 
 	if( d->mTimestamp.isValid() && urlTimestamp == d->mTimestamp ) {
 		// We have the image in cache
-		QCString format;
+		LOG(d->mURL << ", We have the image in cache");
 		d->mRawData = Cache::instance()->file( d->mURL );
-		ImageFrames frames;
-		Cache::instance()->getFrames( d->mURL, frames, format );
-		if( !frames.isEmpty()) {
-			d->mImageFormat = format;
-			d->mFrames = frames;
-			if( !d->mRawData.isNull() || format != "JPEG" ) {
-				finish( true );
+		Cache::instance()->getFrames(d->mURL, &d->mFrames, &d->mImageFormat);
+
+		if( !d->mFrames.isEmpty()) {
+			LOG("The image in cache can be used");
+			d->mProcessedImage = d->mFrames[0].image;
+			emit sizeLoaded(d->mProcessedImage.width(), d->mProcessedImage.height());
+			emit imageChanged(d->mProcessedImage.rect());
+			
+			if (d->mRawData.isNull() && d->mImageFormat=="JPEG") {
+				// Raw data is needed for JPEG, wait for it to be downloaded
+				LOG("Wait for raw data to be downloaded");
+				d->mDecodeState = DECODE_CACHED;
+			} else {
+				// We don't care about raw data
+				finish(true);
 				return;
 			}
-			// the raw data is needed for JPEG, so it needs to be loaded
-			// if it's not in the cache
 		} else {
-			// Image in cache is broken, let's try the file
+			// Image in cache is broken
+			LOG("The image in cache cannot be used");
 			if( !d->mRawData.isNull()) {
+				LOG("Using cached raw data");
+				// Raw data is ok, skip get step and decode it
+				d->mGetState = GET_DONE;
 				d->mTimeSinceLastUpdate.start();
 				d->mDecoderTimer.start(0, false);
 				return;
@@ -350,12 +382,12 @@ void ImageLoader::slotStatResult(KIO::Job* job) {
 
 	d->mTimestamp = urlTimestamp;
 	d->mRawData.resize(0);
-	d->mGetPending = true;
+	d->mGetState = GET_PENDING_GET;
 	checkPendingGet();
 }
 
 void ImageLoader::checkPendingGet() {
-	if( d->mSuspended || !d->mGetPending ) return;
+	if( d->mSuspended || d->mGetState != GET_PENDING_GET ) return;
 
 	// Start loading the image
 	KIO::Job* getJob=KIO::get( d->mURL, false, false);
@@ -368,7 +400,7 @@ void ImageLoader::checkPendingGet() {
 		this, SLOT(slotGetResult(KIO::Job*)) );
 
 	d->mTimeSinceLastUpdate.start();
-	d->mGetPending = false;
+	d->mGetState = GET_GETTING;
 }
 	
 
@@ -380,37 +412,44 @@ void ImageLoader::slotGetResult(KIO::Job* job) {
 		return;
 	}
 
-	d->mGetComplete = true;
+	d->mGetState = GET_DONE;
 	
 	// Store raw data in cache
 	// Note: Cache will give high cost to non-JPEG raw data.
 	Cache::instance()->addFile( d->mURL, d->mRawData, d->mTimestamp );
 
-	if( !d->mImageFormat.isNull()) { // image was in cache, but not raw data
-		finish( true );
-		return;
-	}
 
-	// Start the decoder thread if needed
-	if( d->mUseThread ) {
+	switch (d->mDecodeState) {
+	case DECODE_CACHED:
+		// image was in cache, but not raw data
+		finish( true );
+		break;
+
+	case DECODE_PENDING_THREADED_DECODING:
+		// Start the decoder thread if needed
 		startThread();
-	} else { // Finish decoding if needed
-		if( !d->mDecoderTimer.isActive()) d->mDecoderTimer.start(0);
+		break;
+
+	default:
+		// Finish decoding if needed
+		if (!d->mDecoderTimer.isActive()) d->mDecoderTimer.start(0);
 	}
 }
 
 
 void ImageLoader::slotDataReceived(KIO::Job* job, const QByteArray& chunk) {
-	LOG("size: " << chunk.size());
+	LOG2("size: " << chunk.size());
 	if (chunk.size()<=0) return;
 
 	int oldSize=d->mRawData.size();
 	d->mRawData.resize(oldSize + chunk.size());
 	memcpy(d->mRawData.data()+oldSize, chunk.data(), chunk.size() );
 
-	// Try to determine the data type
 	if (oldSize==0) {
-		d->mURLKind=MimeTypeUtils::determineKindFromContent(d->mRawData);
+		// Try to determine the data type
+		KMimeType::Ptr ptr = KMimeType::findByContent(d->mRawData);
+		d->mMimeType = ptr->name();
+		d->mURLKind = MimeTypeUtils::mimeTypeKind(d->mMimeType);
 		if (d->mURLKind!=MimeTypeUtils::KIND_RASTER_IMAGE) {
 			Q_ASSERT(!d->mDecoderTimer.isActive());
 			job->kill(true /* quietly */);
@@ -421,9 +460,11 @@ void ImageLoader::slotDataReceived(KIO::Job* job, const QByteArray& chunk) {
 		LOG("emit urlKindDetermined(raster)");
 		emit urlKindDetermined();
 	}
-
+	
 	// Decode the received data
-	if( !d->mDecoderTimer.isActive() && !d->mUseThread) {
+	if( !d->mDecoderTimer.isActive() && 
+		(d->mDecodeState==DECODE_WAITING || d->mDecodeState==DECODE_INCREMENTAL_DECODING)
+	) {
 		d->mDecoderTimer.start(0);
 	}
 }
@@ -435,30 +476,29 @@ void ImageLoader::decodeChunk() {
 		d->mDecoderTimer.stop();
 		return;
 	}
-	if( !d->mImageFormat.isNull()) { // image was in cache, only loading the raw data
-		LOG("mImageFormat is not null");
-		d->mDecoderTimer.stop();
-		return;
-	}
 
 	int chunkSize = QMIN(DECODE_CHUNK_SIZE, int(d->mRawData.size())-d->mDecodedSize);
-	LOG("chunkSize: " << chunkSize);
 	int decodedSize = 0;
 	if (chunkSize>0) {
 		decodedSize = d->mDecoder.decode(
 			(const uchar*)(d->mRawData.data()+d->mDecodedSize),
 			chunkSize);
-		LOG("decodedSize: " << decodedSize);
 
 		if (decodedSize<0) {
-			// We can't use async decoding, switch to decoder thread 
+			// We can't use incremental decoding, switch to threaded decoding
 			d->mDecoderTimer.stop();
-			d->mUseThread=true;
-			if( d->mGetComplete ) startThread();	
+			if (d->mGetState == GET_DONE) {
+				startThread();
+			} else {
+				d->mDecodeState = DECODE_PENDING_THREADED_DECODING;
+			}
 			return;
 		}
 
 		// We just decoded some data
+		if (d->mDecodeState == DECODE_WAITING) {
+			d->mDecodeState = DECODE_INCREMENTAL_DECODING;
+		}
 		d->mDecodedSize+=decodedSize;
 	}
 
@@ -466,10 +506,20 @@ void ImageLoader::decodeChunk() {
 		// We decoded as much as possible from the buffer, wait to receive
 		// more data before coming again in decodeChunk
 		d->mDecoderTimer.stop();
-		if (d->mGetComplete && !d->mAsyncImageComplete) {
-			// No more data is available, the image must be truncated,
-			// let's simulate its end
-			end();
+
+		if (d->mGetState == GET_DONE) {
+			// All available data has been received.
+			if (d->mDecodeState == DECODE_INCREMENTAL_DECODING) {
+				// Decoder is not finished, the image must be truncated,
+				// let's simulate its end
+				kdWarning() << "ImageLoader::decodeChunk(): image '" << d->mURL.prettyURL() << "' is truncated.\n";
+
+				if (d->mProcessedImage.isNull()) {
+					d->mProcessedImage = d->mDecoder.image();
+				}
+				emit imageChanged(d->mProcessedImage.rect());
+				end();
+			}
 		}
 	}
 }
@@ -477,6 +527,7 @@ void ImageLoader::decodeChunk() {
 
 void ImageLoader::startThread() {
 	LOG("starting decoder thread");
+	d->mDecodeState = DECODE_THREADED_DECODING;
 	d->mDecoderThread.setRawData(d->mRawData);
 	d->mDecoderThread.start();
 }
@@ -489,36 +540,25 @@ void ImageLoader::slotDecoderThreadFailed() {
 }
 
 
-void ImageLoader::slotImageDecoded() {
+void ImageLoader::slotDecoderThreadSucceeded() {
 	LOG("");
-
-	// Get image
-	if (d->mUseThread) {
-		d->mFrames.clear();
-		d->mFrames.append( ImageFrame( d->mDecoderThread.popLoadedImage(), 0 ));
-	} else if( d->mFrames.count() == 0 ) {
-		d->mFrames.append( ImageFrame( d->mDecoder.image(), 0 ));
-	}
-	
-	// Set image format
-	QBuffer buffer(d->mRawData);
-	buffer.open(IO_ReadOnly);
-	d->mImageFormat = QImageIO::imageFormat(&buffer);
-	buffer.close();
-
-	finish( true );
+	d->mProcessedImage = d->mDecoderThread.popLoadedImage();
+	d->mFrames.append( ImageFrame( d->mProcessedImage, 0 ));
+	emit sizeLoaded(d->mProcessedImage.width(), d->mProcessedImage.height());
+	emit imageChanged(d->mProcessedImage.rect());
+	finish(true);
 }
 
 
 /**
- * Make the final adjustments to the image.
+ * Cache image and emit imageLoaded
  */
 void ImageLoader::finish( bool ok ) {
 	LOG("");
 
-	d->mDecodeComplete = true;
+	d->mDecodeState = DECODE_DONE;
 
-	if( !ok || d->mFrames.count() == 0 ) {
+	if (!ok) {
 		d->mFrames.clear();
 		d->mRawData = QByteArray();
 		d->mImageFormat = QCString();
@@ -526,23 +566,12 @@ void ImageLoader::finish( bool ok ) {
 		emit imageLoaded( false );
 		return;
 	}
-
+	
+	if (d->mImageFormat.isEmpty()) {
+		d->determineImageFormat();
+	}
+	Q_ASSERT(d->mFrames.count() > 0);
 	Cache::instance()->addImage( d->mURL, d->mFrames, d->mImageFormat, d->mTimestamp );
-
-	ImageFrame lastframe = d->mFrames.last();
-	d->mFrames.pop_back(); // maintain that processedImage() is not included when calling imageChanged()
-	d->mProcessedImage = lastframe.image;
-	// The decoder did not cause some signals to be emitted, let's do it now
-	if (d->mKnownSize.isEmpty()) {
-		emit sizeLoaded( lastframe.image.width(), lastframe.image.height());
-	}
-	if (!d->mLoadChangedRect.isEmpty()) {
-		emit imageChanged( d->mLoadChangedRect );
-	} else if (!d->mUpdatedDuringLoad) {
-		emit imageChanged( QRect(QPoint(0,0), lastframe.image.size()) );
-	}
-	d->mFrames.push_back( lastframe );
-
 	emit imageLoaded( true );
 }
 
@@ -550,9 +579,9 @@ void ImageLoader::finish( bool ok ) {
 BusyLevel ImageLoader::priority() const {
 	BusyLevel mylevel = BUSY_NONE;
 	for( QValueVector< OwnerData >::ConstIterator it = d->mOwners.begin();
-	     it != d->mOwners.end();
-	     ++it ) {
-            mylevel = QMAX( mylevel, (*it).priority );
+			it != d->mOwners.end();
+			++it ) {
+		mylevel = QMAX( mylevel, (*it).priority );
 	}
 	return mylevel;
 }
@@ -589,25 +618,101 @@ void ImageLoader::resumeLoading() {
 void ImageLoader::end() {
 	LOG("");
 
+	// Notify about the last loaded rectangle
+	LOG("mLoadChangedRect " << d->mLoadChangedRect);
+	if (!d->mLoadChangedRect.isEmpty()) {
+		emit imageChanged( d->mLoadChangedRect );
+	}
+
 	d->mDecoderTimer.stop();
-	d->mAsyncImageComplete=true;
-	
+	d->mDecodeState = DECODE_INCREMENTAL_DECODING_DONE;
+
+	// We are done
+	if( d->mFrames.count() == 0 ) {
+		d->mFrames.append( ImageFrame( d->mProcessedImage, 0 ));
+	}
 	// The image has been totally decoded, we delay the call to finish because
 	// when we return from this function we will be in decodeChunk(), after the
 	// call to decode(), so we don't want to switch to a new impl yet, since
 	// this means deleting "this".
-	QTimer::singleShot(0, this, SLOT(slotImageDecoded()) );
+	QTimer::singleShot(0, this, SLOT(callFinish()) );
 }
 
-void ImageLoader::changed(const QRect& rect) {
-	d->mProcessedImage = d->mDecoder.image();
+
+void ImageLoader::callFinish() {
+	finish(true);
+}
+
+
+void ImageLoader::changed(const QRect& constRect) {
+	LOG2("");
+	QRect rect = constRect;
+	
+	if (d->mLoadedRegion.isEmpty()) {
+		// This is the first time we get called. Init mProcessedImage and emit
+		// sizeLoaded.
+		LOG("mLoadedRegion is empty");
+		
+		// By default, mProcessedImage should use the image from mDecoder
+		d->mProcessedImage = d->mDecoder.image();
+		
+		if (d->mImageFormat.isEmpty()) {
+			d->determineImageFormat();
+		}
+		Q_ASSERT(!d->mImageFormat.isEmpty());
+		if (d->mImageFormat == "JPEG") {
+			// This is a JPEG, extract orientation and adjust mProcessedImage
+			// if necessary
+			ImageUtils::JPEGContent content;
+
+			if (content.loadFromData(d->mRawData)) {
+				d->mOrientation = content.orientation();
+				if (d->mOrientation != ImageUtils::NOT_AVAILABLE && d->mOrientation != ImageUtils::NORMAL) {
+					QSize size = content.size();
+					d->mProcessedImage = QImage(size, d->mDecoder.image().depth());
+				}
+			} else {
+				kdWarning() << "ImageLoader::changed(): JPEGContent could not load '" << d->mURL.prettyURL() << "'\n";
+			}
+		}
+		
+		LOG("emit sizeLoaded " << d->mProcessedImage.size());
+		emit sizeLoaded(d->mProcessedImage.width(), d->mProcessedImage.height());
+	}
+
+	// Apply orientation if necessary
+	if (d->mOrientation != ImageUtils::NOT_AVAILABLE && d->mOrientation != ImageUtils::NORMAL) {
+		// We can only rotate whole images, so copy the loaded rect in a temp
+		// image, rotate the temp image and copy it to mProcessedImage
+
+		// Copy loaded rect
+		QImage temp(rect.size(), d->mProcessedImage.depth());
+		bitBlt(&temp, 0, 0, 
+			&d->mDecoder.image(), rect.left(), rect.top(), rect.width(), rect.height());
+		
+		// Rotate
+		temp = ImageUtils::transform(temp, d->mOrientation);
+
+		// Compute destination rect
+		QWMatrix matrix = ImageUtils::transformMatrix(d->mOrientation);
+		
+		QRect imageRect = d->mDecoder.image().rect();
+		imageRect = matrix.mapRect(imageRect);
+				
+		rect = matrix.mapRect(rect);
+		rect.moveBy(-imageRect.left(), -imageRect.top());
+		
+		// copy temp to mProcessedImage
+		bitBlt(&d->mProcessedImage, rect.left(), rect.top(),
+			&temp, 0, 0, temp.width(), temp.height());
+	}
+	
+	// Update state tracking vars
 	d->mWasFrameData = true;
-	d->mUpdatedDuringLoad=true;
 	d->mLoadChangedRect |= rect;
 	d->mLoadedRegion |= rect;
-	if( d->mTimeSinceLastUpdate.elapsed() > 100 ) {
-		LOG(d->mLoadChangedRect.left() << "-" << d->mLoadChangedRect.top()
-			<< " " << d->mLoadChangedRect.width() << "x" << d->mLoadChangedRect.height() );
+	if( d->mTimeSinceLastUpdate.elapsed() > IMAGE_UPDATE_INTERVAL ) {
+		LOG("emitting imageChanged " << d->mLoadChangedRect);
 		d->mTimeSinceLastUpdate.start();
 		emit imageChanged(d->mLoadChangedRect);
 		d->mLoadChangedRect = QRect();
@@ -615,10 +720,11 @@ void ImageLoader::changed(const QRect& rect) {
 }
 
 void ImageLoader::frameDone() {
-	frameDone( QPoint( 0, 0 ), QRect( 0, 0, d->mDecoder.image().width(), d->mDecoder.image().height()));
+	frameDone( QPoint( 0, 0 ), d->mDecoder.image().rect());
 }
 
 void ImageLoader::frameDone(const QPoint& offset, const QRect& rect) {
+	LOG("");
 	// Another case where the image loading in Qt's is a bit borken.
 	// It's possible to get several notes about a frame being done for one frame (with MNG).
 	if( !d->mWasFrameData ) {
@@ -642,19 +748,24 @@ void ImageLoader::frameDone(const QPoint& offset, const QRect& rect) {
 		d->mTimeSinceLastUpdate.start();
 	}
 	d->mLoadedRegion = QRegion();
-	QImage image = d->mDecoder.image();
-	image.detach();
-	if( offset != QPoint( 0, 0 ) || rect != QRect( 0, 0, image.width(), image.height())) {
-                if( !d->mFrames.isEmpty()) { // not first image - what to do here?
-			QImage im( d->mFrames.last().image);
-			im.detach();
+	
+	QImage image;
+	if (d->mProcessedImage.isNull()) {
+		image = d->mDecoder.image().copy();
+	} else {
+		image = d->mProcessedImage.copy();
+	}
+	
+	if( offset != QPoint( 0, 0 ) || rect != image.rect()) {
+		// Blit last frame below 'image'
+		if( !d->mFrames.isEmpty()) {
+			QImage im = d->mFrames.last().image.copy();
 			bitBlt( &im, offset.x(), offset.y(), &image, rect.x(), rect.y(), rect.width(), rect.height());
 			image = im;
 		}
 	}
 	d->mFrames.append( ImageFrame( image, d->mNextFrameDelay ));
 	d->mNextFrameDelay = 0;
-	emit frameLoaded();
 }
 
 void ImageLoader::setLooping(int) {
@@ -667,12 +778,9 @@ void ImageLoader::setFramePeriod(int milliseconds) {
 	}
 }
 
-void ImageLoader::setSize(int width, int height) {
-	LOG(width << "x" << height);
-	d->mKnownSize = QSize( width, height );
-	// FIXME: There must be a better way than creating an empty image
-	d->mProcessedImage = QImage( width, height, 32 );
-	emit sizeLoaded(width, height);
+void ImageLoader::setSize(int, int) {
+	// Do nothing, size is handled when ::changed() is called for the first
+	// time
 }
 
 
@@ -681,11 +789,6 @@ QImage ImageLoader::processedImage() const {
 }
 
 
-QSize ImageLoader::knownSize() const {
-	return d->mKnownSize;
-}
-
-
 ImageFrames ImageLoader::frames() const {
 	return d->mFrames;
 }
@@ -701,6 +804,11 @@ QByteArray ImageLoader::rawData() const {
 }
 
 
+QString ImageLoader::mimeType() const {
+	return d->mMimeType;
+}
+
+
 MimeTypeUtils::Kind ImageLoader::urlKind() const {
 	return d->mURLKind;
 }
@@ -717,7 +825,7 @@ QRegion ImageLoader::loadedRegion() const {
 
 
 bool ImageLoader::completed() const {
-	return d->mDecodeComplete;
+	return d->mDecodeState == DECODE_DONE;
 }
 
 
@@ -736,7 +844,7 @@ void ImageLoader::deref( const QObject* owner ) {
 		if( (*it).owner == owner ) {
 			d->mOwners.erase( it );
 			if( d->mOwners.count() == 0 ) {
-				loaders.remove( d->mURL );
+				sLoaders.remove( d->mURL );
 				delete this;
 			}
 			return;
@@ -761,24 +869,24 @@ void ImageLoader::ownerDestroyed() {
 //---------------------------------------------------------------------
 
 ImageLoader* ImageLoader::loader( const KURL& url, const QObject* owner, BusyLevel priority ) {
-	if( loaders.contains( url )) {
-		ImageLoader* l = loaders[ url ];
-		l->ref( owner, priority );
+	if( sLoaders.contains( url )) {
+		ImageLoader* loader = sLoaders[ url ];
+		loader->ref( owner, priority );
 		// resume if this owner has high priority
-		l->slotBusyLevelChanged( BusyLevelManager::instance()->busyLevel());
-		return l;
+		loader->slotBusyLevelChanged( BusyLevelManager::instance()->busyLevel());
+		return loader;
 	}
-	ImageLoader* l = new ImageLoader;
-	l->ref( owner, priority );
-	loaders[ url ] = l;
-	l->setURL( url );
+	ImageLoader* loader = new ImageLoader;
+	loader->ref( owner, priority );
+	sLoaders[ url ] = loader;
+	loader->setURL( url );
 	// Code using a loader first calls loader() to get ImageLoader* and only after that it can
 	// connect to its signals etc., so don't start loading immediately.
 	// This also helps with preloading jobs, since BUSY_LOADING busy level is not entered immediately
 	// when a new picture is selected, so preloading jobs without this delay could start working
 	// immediately.
-	QTimer::singleShot( priority >= BUSY_LOADING ? 0 : 10, l, SLOT( startLoading()));
-	return l;
+	QTimer::singleShot( priority >= BUSY_LOADING ? 0 : 10, loader, SLOT( startLoading()));
+	return loader;
 }
 
 } // namespace
