@@ -20,19 +20,24 @@ Foundation, Inc., 51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.
 
 #include <qapplication.h>
 #include <qcursor.h>
+#include <qfile.h>
 #include <qpoint.h>
 
 #include <kaction.h>
+#include <kapplication.h>
 #include <kconfig.h>
 #include <kdebug.h>
 #include <kdirlister.h>
+#include <kfiledialog.h>
 #include <kiconloader.h>
 #include <klocale.h>
+#include <kmessagebox.h>
 #include <kmimetype.h>
 #include <kparts/genericfactory.h>
 
 #include <gvcore/cache.h>
 #include <gvcore/document.h>
+#include <gvcore/fileoperation.h>
 #include <gvcore/printdialog.h>
 #include <gvcore/imageview.h>
 #include <gvcore/imageloader.h>
@@ -46,7 +51,7 @@ const char CONFIG_CACHE_GROUP[]="cache";
 
 #undef ENABLE_LOG
 #undef LOG
-//#define ENABLE_LOG
+#define ENABLE_LOG
 #ifdef ENABLE_LOG
 #define LOG(x) kdDebug() << k_funcinfo << x << endl
 #else
@@ -54,6 +59,18 @@ const char CONFIG_CACHE_GROUP[]="cache";
 #endif
 
 
+static bool storeData(QWidget* parent, QFile* file, const QByteArray& data) {
+	uint sizeWritten = file->writeBlock(data);
+	if (sizeWritten != data.size()) {
+		KMessageBox::error(
+			parent,
+			i18n("Could not save image to a temporary file"));
+		return false;
+	}
+	return true;
+}
+
+
 //Factory Code
 typedef KParts::GenericFactory<GVImagePart> GVImageFactory;
 K_EXPORT_COMPONENT_FACTORY( libgvimagepart /*library name*/, GVImageFactory )
@@ -67,8 +84,6 @@ GVImagePart::GVImagePart(QWidget* parentWidget, const char* /*widgetName*/, QObj
 	setInstance( GVImageFactory::instance() );
 	KGlobal::locale()->insertCatalogue( "gwenview" );
 
-	Cache::instance()->ref();
-
 	mBrowserExtension = new GVImagePartBrowserExtension(this);
 
 	// Create the widgets
@@ -77,12 +92,12 @@ GVImagePart::GVImagePart(QWidget* parentWidget, const char* /*widgetName*/, QObj
 	connect( mDocument, SIGNAL( loaded(const KURL&)), SLOT( slotLoaded(const KURL&)));
 	mImageView = new ImageView(parentWidget, mDocument, actionCollection());
 	connect( mImageView, SIGNAL(requestContextMenu(const QPoint&)),
-		mBrowserExtension, SLOT(openContextMenu(const QPoint&)) );
+		this, SLOT(openContextMenu(const QPoint&)) );
 	setWidget(mImageView);
 
 	mDirLister = new KDirLister;
 	mDirLister->setAutoErrorHandlingEnabled( false, 0 );
-	mDirLister->setMainWindow( parentWidget->topLevelWidget());
+	mDirLister->setMainWindow(KApplication::kApplication()->mainWidget());
 	connect( mDirLister, SIGNAL( clear()), SLOT( dirListerClear()));
 	connect( mDirLister, SIGNAL( newItems( const KFileItemList& )),
 		SLOT( dirListerNewItems( const KFileItemList& )));
@@ -99,14 +114,14 @@ GVImagePart::GVImagePart(QWidget* parentWidget, const char* /*widgetName*/, QObj
 		this,SLOT(slotSelectNext()), actionCollection(), "next");
 	updateNextPrevious();
 
-	KStdAction::saveAs( mDocument, SLOT(saveAs()), actionCollection(), "saveAs" );
+	KStdAction::saveAs( this, SLOT(saveAs()), actionCollection(), "saveAs" );
+	new KAction(i18n("Rotate &Left"), "rotate_ccw", CTRL + Key_L, this, SLOT(rotateLeft()), actionCollection(), "rotate_left");
 	new KAction(i18n("Rotate &Right"), "rotate_cw", CTRL + Key_R, this, SLOT(rotateRight()), actionCollection(), "rotate_right");
 
 	setXMLFile( "gvimagepart/gvimagepart.rc" );
 }
 
 GVImagePart::~GVImagePart() {
-	Cache::instance()->deref();
 	delete mDirLister;
 }
 
@@ -121,6 +136,19 @@ void GVImagePart::partActivateEvent(KParts::PartActivateEvent* event) {
 }
 
 
+void GVImagePart::guiActivateEvent( KParts::GUIActivateEvent* event) {
+	// Stolen from KHTMLImage
+	//
+	// prevent the base implementation from emitting setWindowCaption with
+	// our url. It destroys our pretty, previously caption. Konq saves/restores
+	// the caption for us anyway.
+	if (event->activated()) {
+		return;
+	}
+	KParts::ReadOnlyPart::guiActivateEvent(event);
+}
+
+
 KAboutData* GVImagePart::createAboutData() {
 	KAboutData* aboutData = new KAboutData( "gvimagepart", I18N_NOOP("GVImagePart"),
 						"0.1", I18N_NOOP("Image Viewer"),
@@ -193,6 +221,10 @@ void GVImagePart::print() {
 	}
 }
 
+void GVImagePart::rotateLeft() {
+	mDocument->transform(ImageUtils::ROT_270);
+}
+
 void GVImagePart::rotateRight() {
 	mDocument->transform(ImageUtils::ROT_90);
 }
@@ -271,6 +303,134 @@ void GVImagePart::slotSelectPrevious() {
 	emit mBrowserExtension->openURLNotify();
 }
 
+
+void GVImagePart::saveAs() {
+	if (!mDocument->isModified()) {
+		saveOriginalAs();
+		return;
+	}
+
+	if (mDocument->canBeSaved()) {
+		mDocument->saveAs();
+		return;
+	}
+
+	KGuiItem saveItem(i18n("&Save Original"), "filesaveas");
+	int result = KMessageBox::warningContinueCancel(
+		widget(), 
+		i18n("Gwenview KPart can't save the modifications you made. Do you want to save the original image?"),
+		i18n("Warning"),
+		saveItem);
+
+	if (result == KMessageBox::Cancel) return;
+
+	saveOriginalAs();
+}
+
+
+void GVImagePart::showJobError(KIO::Job* job) {
+	if (job->error() != 0) { 
+		job->showErrorDialog(widget());
+	}
+}
+
+
+void GVImagePart::saveOriginalAs() {
+	KURL srcURL = mDocument->url();
+	KURL dstURL = KFileDialog::getSaveURL(
+		srcURL.fileName(),
+		QString::null,
+		widget());
+	if (!dstURL.isValid()) return;
+
+	// Try to get data from the cache to avoid downloading the image again.
+	QByteArray data = Cache::instance()->file(srcURL);
+
+	if (data.size() == 0) {
+		// We need to read the image again. Let KIO::copy do the work.
+		KIO::Job* job = KIO::copy(srcURL, dstURL);
+		job->setWindow(widget());
+		connect(job, SIGNAL(result(KIO::Job*)),
+			this, SLOT(showJobError(KIO::Job*)) );
+		return;
+	}
+
+	if (dstURL.isLocalFile()) {
+		// Destination is a local file, store it ourself
+		QString path = dstURL.path();
+		QFile file(path);
+		if (!file.open(IO_WriteOnly)) {
+			KMessageBox::error(
+				widget(),
+				i18n("Could not open '%1' for writing.").arg(path));
+			return;
+		}
+		storeData(widget(), &file, data);
+		return;
+	}
+
+	// We need to send the data to a remote location
+	new DataUploader(widget(), data, dstURL);
+}
+
+
+DataUploader::DataUploader(QWidget* dialogParent, const QByteArray& data, const KURL& dstURL)
+: mDialogParent(dialogParent)
+{
+	mTempFile.setAutoDelete(true);
+
+	// Store it in a temp file
+	if (! storeData(dialogParent, mTempFile.file(), data) ) return;
+
+	// Now upload it
+	KURL tmpURL;
+	tmpURL.setPath(mTempFile.name());
+	KIO::Job* job = KIO::copy(tmpURL, dstURL);
+	job->setWindow(dialogParent);
+	connect(job, SIGNAL(result(KIO::Job*)),
+		this, SLOT(slotJobFinished(KIO::Job*)) );
+}
+
+
+void DataUploader::slotJobFinished(KIO::Job* job) {
+	if (job->error() != 0) { 
+		job->showErrorDialog(mDialogParent);
+	}
+
+	delete this;
+}
+
+
+/**
+ * Overload KXMLGUIClient so that we can call setXML
+ */
+class PopupGUIClient : public KXMLGUIClient {
+public:
+	PopupGUIClient( KInstance *inst, const QString &doc ) {
+		setInstance( inst );
+		setXML( doc );
+	}
+};
+
+
+void GVImagePart::openContextMenu(const QPoint& pos) {
+	QString doc = KXMLGUIFactory::readConfigFile( "gvimagepartpopup.rc", true, instance() );
+	PopupGUIClient guiClient(instance(), doc);
+	
+	KStdAction::saveAs( this, SLOT(saveAs()), guiClient.actionCollection(), "saveAs" );
+	
+	KParts::URLArgs urlArgs;
+	urlArgs.serviceType = mDocument->mimeType();
+
+	KParts::BrowserExtension::PopupFlags flags = 
+		KParts::BrowserExtension::ShowNavigationItems 
+		| KParts::BrowserExtension::ShowUp
+		| KParts::BrowserExtension::ShowReload;
+	
+	emit mBrowserExtension->popupMenu(&guiClient, pos, m_url, urlArgs, flags, S_IFREG);
+}
+
+
 /***** GVImagePartBrowserExtension *****/
 
 GVImagePartBrowserExtension::GVImagePartBrowserExtension(GVImagePart* viewPart, const char* name)
@@ -282,15 +442,6 @@ GVImagePartBrowserExtension::GVImagePartBrowserExtension(GVImagePart* viewPart,
 GVImagePartBrowserExtension::~GVImagePartBrowserExtension() {
 }
 
-void GVImagePartBrowserExtension::openContextMenu(const QPoint& pos) {
-	KURL url=mGVImagePart->url();
-	QString mimeType=KMimeType::findByURL(url)->name();
-	KFileItem item(url, mimeType, S_IFREG);
-	KFileItemList list;
-	list.append(&item);
-	emit popupMenu(pos, list);
-}
-
 void GVImagePartBrowserExtension::print() {
 	mGVImagePart->print();
 }
